@@ -1,9 +1,11 @@
+# frozen_string_literal: true
 ##
 # Parses a gem.deps.rb.lock file and constructs a LockSet containing the
 # dependencies found inside.  If the lock file is missing no LockSet is
 # constructed.
 
 class Gem::RequestSet::Lockfile
+
   ##
   # Raised when a lockfile cannot be parsed
 
@@ -28,25 +30,26 @@ class Gem::RequestSet::Lockfile
     # Raises a ParseError with the given +message+ which was encountered at a
     # +line+ and +column+ while parsing.
 
-    def initialize message, column, line, path
+    def initialize(message, column, line, path)
       @line   = line
       @column = column
       @path   = path
       super "#{message} (at line #{line} column #{column})"
     end
+
   end
 
   ##
   # Creates a new Lockfile for the given +request_set+ and +gem_deps_file+
   # location.
 
-  def self.build request_set, gem_deps_file, dependencies = nil
+  def self.build(request_set, gem_deps_file, dependencies = nil)
     request_set.resolve
     dependencies ||= requests_to_deps request_set.sorted_requests
     new request_set, gem_deps_file, dependencies
   end
 
-  def self.requests_to_deps requests # :nodoc:
+  def self.requests_to_deps(requests) # :nodoc:
     deps = {}
 
     requests.each do |request|
@@ -55,7 +58,7 @@ class Gem::RequestSet::Lockfile
       requirement = request.request.dependency.requirement
 
       deps[name] = if [Gem::Resolver::VendorSpecification,
-                       Gem::Resolver::GitSpecification].include? spec.class then
+                       Gem::Resolver::GitSpecification].include? spec.class
                      Gem::Requirement.source_set
                    else
                      requirement
@@ -70,7 +73,7 @@ class Gem::RequestSet::Lockfile
 
   attr_reader :platforms
 
-  def initialize request_set, gem_deps_file, dependencies
+  def initialize(request_set, gem_deps_file, dependencies)
     @set           = request_set
     @dependencies  = dependencies
     @gem_deps_file = File.expand_path(gem_deps_file)
@@ -78,10 +81,10 @@ class Gem::RequestSet::Lockfile
 
     @gem_deps_file.untaint unless gem_deps_file.tainted?
 
-    @platforms      = []
+    @platforms = []
   end
 
-  def add_DEPENDENCIES out # :nodoc:
+  def add_DEPENDENCIES(out) # :nodoc:
     out << "DEPENDENCIES"
 
     out.concat @dependencies.sort_by { |name,| name }.map { |name, requirement|
@@ -91,7 +94,7 @@ class Gem::RequestSet::Lockfile
     out << nil
   end
 
-  def add_GEM out, spec_groups # :nodoc:
+  def add_GEM(out, spec_groups) # :nodoc:
     return if spec_groups.empty?
 
     source_groups = spec_groups.values.flatten.group_by do |request|
@@ -121,7 +124,7 @@ class Gem::RequestSet::Lockfile
     end
   end
 
-  def add_GIT out, git_requests
+  def add_GIT(out, git_requests)
     return if git_requests.empty?
 
     by_repository_revision = git_requests.group_by do |request|
@@ -129,8 +132,8 @@ class Gem::RequestSet::Lockfile
       [source.repository, source.rev_parse]
     end
 
-    out << "GIT"
     by_repository_revision.each do |(repository, revision), requests|
+      out << "GIT"
       out << "  remote: #{repository}"
       out << "  revision: #{revision}"
       out << "  specs:"
@@ -143,17 +146,16 @@ class Gem::RequestSet::Lockfile
           out << "      #{dep.name}#{dep.requirement.for_lockfile}"
         end
       end
+      out << nil
     end
-
-    out << nil
   end
 
-  def relative_path_from dest, base # :nodoc:
+  def relative_path_from(dest, base) # :nodoc:
     dest = File.expand_path(dest)
     base = File.expand_path(base)
 
-    if dest.index(base) == 0 then
-      offset = dest[base.size+1..-1]
+    if dest.index(base) == 0
+      offset = dest[base.size + 1..-1]
 
       return '.' unless offset
 
@@ -163,7 +165,7 @@ class Gem::RequestSet::Lockfile
     end
   end
 
-  def add_PATH out, path_requests # :nodoc:
+  def add_PATH(out, path_requests) # :nodoc:
     return if path_requests.empty?
 
     out << "PATH"
@@ -178,7 +180,7 @@ class Gem::RequestSet::Lockfile
     out << nil
   end
 
-  def add_PLATFORMS out # :nodoc:
+  def add_PLATFORMS(out) # :nodoc:
     out << "PLATFORMS"
 
     platforms = requests.map { |request| request.spec.platform }.uniq
@@ -223,7 +225,7 @@ class Gem::RequestSet::Lockfile
   def write
     content = to_s
 
-    open "#{@gem_deps_file}.lock", 'w' do |io|
+    File.open "#{@gem_deps_file}.lock", 'w' do |io|
       io.write content
     end
   end
@@ -233,6 +235,7 @@ class Gem::RequestSet::Lockfile
   def requests
     @set.sorted_requests
   end
+
 end
 
 require 'rubygems/request_set/lockfile/tokenizer'
