@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 #
 # test_scanner_events.rb
 #
@@ -14,13 +15,23 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
   def test_event_coverage
     dispatched = Ripper::SCANNER_EVENTS.map {|event,_| event }
     dispatched.each do |e|
-      assert_equal true, respond_to?("test_#{e}", true), "event not tested: #{e}"
+      assert_respond_to(self, ["test_#{e}", true], "event not tested: #{e}")
     end
   end
 
-  def scan(target, str)
+  def scan(target, str, &error)
     sym = "on_#{target}".intern
-    Ripper.lex(str).select {|_1,type,_2| type == sym }.map {|_1,_2,tok| tok }
+    lexer = Ripper::Lexer.new(str)
+    if error
+      lexer.singleton_class.class_eval do
+        define_method(:on_error) {|ev|
+          yield __callee__, ev, token()
+        }
+        alias on_parse_error on_error
+        alias compile_error on_error
+      end
+    end
+    lexer.lex.select {|_1,type,_2| type == sym }.map {|_1,_2,tok| tok }
   end
 
   def test_tokenize
@@ -32,87 +43,114 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  Ripper.tokenize('1')
     assert_equal ['1', ';', 'def', ' ', 'm', '(', 'arg', ')', 'end'],
                  Ripper.tokenize("1;def m(arg)end")
-    assert_equal ['print', '(', '<<EOS', ')', "\n", "heredoc\n", "EOS\n"],
-                 Ripper.tokenize("print(<<EOS)\nheredoc\nEOS\n")
-    assert_equal ['print', '(', ' ', '<<EOS', ')', "\n", "heredoc\n", "EOS\n"],
-                 Ripper.tokenize("print( <<EOS)\nheredoc\nEOS\n")
+    assert_equal ['print', '(', '<<''EOS', ')', "\n", "heredoc\n", "EOS\n"],
+                 Ripper.tokenize("print(<<""EOS)\nheredoc\nEOS\n")
+    assert_equal ['print', '(', ' ', '<<''EOS', ')', "\n", "heredoc\n", "EOS\n"],
+                 Ripper.tokenize("print( <<""EOS)\nheredoc\nEOS\n")
     assert_equal ["\#\n", "\n", "\#\n", "\n", "nil", "\n"],
                  Ripper.tokenize("\#\n\n\#\n\nnil\n")
+    assert_equal ["1", "  ", ".", "foo", "\n"],
+                 Ripper.tokenize("1  .foo\n")
+    assert_equal ["1", "\n", "  ", ".", "foo", "\n"],
+                 Ripper.tokenize("1\n  .foo\n")
   end
 
   def test_lex
     assert_equal [],
                  Ripper.lex('')
-    assert_equal [[[1,0], :on_ident, "a"]],
+    assert_equal [[[1,0], :on_ident, "a", Ripper::EXPR_CMDARG]],
                  Ripper.lex('a')
-    assert_equal [[[1, 0], :on_kw, "nil"]],
+    assert_equal [[[1, 0], :on_kw, "nil", Ripper::EXPR_END]],
                  Ripper.lex("nil")
-    assert_equal [[[1, 0], :on_kw, "def"],
-                  [[1, 3], :on_sp, " "],
-                  [[1, 4], :on_ident, "m"],
-                  [[1, 5], :on_lparen, "("],
-                  [[1, 6], :on_ident, "a"],
-                  [[1, 7], :on_rparen, ")"],
-                  [[1, 8], :on_kw, "end"]],
+    assert_equal [[[1, 0], :on_kw, "def", Ripper::EXPR_FNAME],
+                  [[1, 3], :on_sp, " ", Ripper::EXPR_FNAME],
+                  [[1, 4], :on_ident, "m", Ripper::EXPR_ENDFN],
+                  [[1, 5], :on_lparen, "(", Ripper::EXPR_BEG | Ripper::EXPR_LABEL],
+                  [[1, 6], :on_ident, "a", Ripper::EXPR_ARG],
+                  [[1, 7], :on_rparen, ")", Ripper::EXPR_ENDFN],
+                  [[1, 8], :on_kw, "end", Ripper::EXPR_END]],
                  Ripper.lex("def m(a)end")
-    assert_equal [[[1, 0], :on_int, "1"],
-                  [[1, 1], :on_nl, "\n"],
-                  [[2, 0], :on_int, "2"],
-                  [[2, 1], :on_nl, "\n"],
-                  [[3, 0], :on_int, "3"]],
+    assert_equal [[[1, 0], :on_int, "1", Ripper::EXPR_END],
+                  [[1, 1], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_int, "2", Ripper::EXPR_END],
+                  [[2, 1], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[3, 0], :on_int, "3", Ripper::EXPR_END]],
                  Ripper.lex("1\n2\n3")
-    assert_equal [[[1, 0], :on_heredoc_beg, "<<EOS"],
-                  [[1, 5], :on_nl, "\n"],
-                  [[2, 0], :on_tstring_content, "heredoc\n"],
-                  [[3, 0], :on_heredoc_end, "EOS"]],
-                 Ripper.lex("<<EOS\nheredoc\nEOS")
-    assert_equal [[[1, 0], :on_heredoc_beg, "<<EOS"],
-                  [[1, 5], :on_nl, "\n"],
-                  [[2, 0], :on_heredoc_end, "EOS"]],
-                 Ripper.lex("<<EOS\nEOS"),
+    assert_equal [[[1, 0], :on_heredoc_beg, "<<""EOS", Ripper::EXPR_BEG],
+                  [[1, 5], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_tstring_content, "heredoc\n", Ripper::EXPR_BEG],
+                  [[3, 0], :on_heredoc_end, "EOS", Ripper::EXPR_BEG]],
+                 Ripper.lex("<<""EOS\nheredoc\nEOS")
+    assert_equal [[[1, 0], :on_heredoc_beg, "<<""EOS", Ripper::EXPR_BEG],
+                  [[1, 5], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_heredoc_end, "EOS", Ripper::EXPR_BEG]],
+                 Ripper.lex("<<""EOS\nEOS"),
                  "bug#4543"
-    assert_equal [[[1, 0], :on_regexp_beg, "/"],
-                  [[1, 1], :on_tstring_content, "foo\nbar"],
-                  [[2, 3], :on_regexp_end, "/"]],
+    assert_equal [[[1, 0], :on_regexp_beg, "/", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo\nbar", Ripper::EXPR_BEG],
+                  [[2, 3], :on_regexp_end, "/", Ripper::EXPR_BEG]],
                  Ripper.lex("/foo\nbar/")
-    assert_equal [[[1, 0], :on_regexp_beg, "/"],
-                  [[1, 1], :on_tstring_content, "foo\n\u3020"],
-                  [[2, 3], :on_regexp_end, "/"]],
+    assert_equal [[[1, 0], :on_regexp_beg, "/", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo\n\u3020", Ripper::EXPR_BEG],
+                  [[2, 3], :on_regexp_end, "/", Ripper::EXPR_BEG]],
                  Ripper.lex("/foo\n\u3020/")
-    assert_equal [[[1, 0], :on_tstring_beg, "'"],
-                  [[1, 1], :on_tstring_content, "foo\n\xe3\x80\xa0"],
-                  [[2, 3], :on_tstring_end, "'"]],
+    assert_equal [[[1, 0], :on_tstring_beg, "'", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo\n\xe3\x80\xa0", Ripper::EXPR_BEG],
+                  [[2, 3], :on_tstring_end, "'", Ripper::EXPR_END]],
                  Ripper.lex("'foo\n\xe3\x80\xa0'")
-    assert_equal [[[1, 0], :on_tstring_beg, "'"],
-                  [[1, 1], :on_tstring_content, "\u3042\n\u3044"],
-                  [[2, 3], :on_tstring_end, "'"]],
+    assert_equal [[[1, 0], :on_tstring_beg, "'", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "\u3042\n\u3044", Ripper::EXPR_BEG],
+                  [[2, 3], :on_tstring_end, "'", Ripper::EXPR_END]],
                  Ripper.lex("'\u3042\n\u3044'")
+    assert_equal [[[1, 0], :on_rational, "1r", Ripper::EXPR_END],
+                  [[1, 2], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_imaginary, "2i", Ripper::EXPR_END],
+                  [[2, 2], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[3, 0], :on_imaginary, "3ri", Ripper::EXPR_END],
+                  [[3, 3], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[4, 0], :on_rational, "4.2r", Ripper::EXPR_END],
+                  [[4, 4], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[5, 0], :on_imaginary, "5.6ri", Ripper::EXPR_END],
+                 ],
+                 Ripper.lex("1r\n2i\n3ri\n4.2r\n5.6ri")
+     assert_equal [[[1, 0], :on_heredoc_beg, "<<~EOS", Ripper::EXPR_BEG],
+                   [[1, 6], :on_nl, "\n", Ripper::EXPR_BEG],
+                   [[2, 0], :on_ignored_sp, "  ", Ripper::EXPR_BEG],
+                   [[2, 2], :on_tstring_content, "heredoc\n", Ripper::EXPR_BEG],
+                   [[3, 0], :on_heredoc_end, "EOS", Ripper::EXPR_BEG]
+                 ],
+                 Ripper.lex("<<~EOS\n  heredoc\nEOS")
+    assert_equal [[[1, 0], :on_tstring_beg, "'", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo", Ripper::EXPR_BEG]],
+                 Ripper.lex("'foo")
   end
 
   def test_location
-    validate_location ""
-    validate_location " "
-    validate_location "@"
-    validate_location "\n"
-    validate_location "\r\n"
-    validate_location "\n\n\n\n\n\r\n\n\n"
-    validate_location "\n;\n;\n;\n;\n"
-    validate_location "nil"
-    validate_location "@ivar"
-    validate_location "1;2;3"
-    validate_location "1\n2\n3"
-    validate_location "1\n2\n3\n"
-    validate_location "def m(a) nil end"
-    validate_location "if true then false else nil end"
-    validate_location "BEGIN{print nil}"
-    validate_location "%w(a b\nc\r\nd \ne )"
-    validate_location %Q["a\nb\r\nc"]
-    validate_location "print(<<EOS)\nheredoc\nEOS\n"
-    validate_location "print(<<-\"EOS\")\nheredoc\n     EOS\n"
+    assert_location ""
+    assert_location " "
+    assert_location ":"
+    assert_location "\n"
+    assert_location "\r\n"
+    assert_location "\n\n\n\n\n\r\n\n\n"
+    assert_location "\n;\n;\n;\n;\n"
+    assert_location "nil"
+    assert_location "@ivar"
+    assert_location "1;2;3"
+    assert_location "1\n2\n3"
+    assert_location "1\n2\n3\n"
+    assert_location "def m(a) nil end"
+    assert_location "if true then false else nil end"
+    assert_location "BEGIN{print nil}"
+    assert_location "%w(a b\nc\r\nd \ne )"
+    assert_location %Q["a\nb\r\nc"]
+    assert_location "print(<<""EOS)\nheredoc\nEOS\n"
+    assert_location "print(<<-\"EOS\")\nheredoc\n     EOS\n"
+    assert_location "'foo'"
+    assert_location "'foo"
   end
 
-  def validate_location(src)
-    buf = ''
+  def assert_location(src)
+    buf = ''.dup
     Ripper.lex(src).each do |pos, type, tok|
       line, col = *pos
       assert_equal buf.count("\n") + 1, line,
@@ -140,7 +178,7 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
     assert_equal [],
                  scan('comma', %q[".,.,.,.,.,.,.."])
     assert_equal [],
-                 scan('comma', "<<EOS\n,,,,,,,,,,\nEOS")
+                 scan('comma', "<<""EOS\n,,,,,,,,,,\nEOS")
   end
 
   def test_period
@@ -206,7 +244,7 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
     assert_equal ['#{'],
                  scan('embexpr_beg', '%Q[#{expr}]')
     assert_equal ['#{'],
-                 scan('embexpr_beg', "m(<<EOS)\n\#{expr}\nEOS")
+                 scan('embexpr_beg', "m(<<""EOS)\n\#{expr}\nEOS")
   end
 
   def test_embexpr_end
@@ -219,7 +257,7 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
     assert_equal ['}'],
                  scan('embexpr_end', '%Q[#{expr}]')
     assert_equal ['}'],
-                 scan('embexpr_end', "m(<<EOS)\n\#{expr}\nEOS")
+                 scan('embexpr_end', "m(<<""EOS)\n\#{expr}\nEOS")
   end
 
   def test_embvar
@@ -231,6 +269,8 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('embvar', '"#@ivar"')
     assert_equal ['#'],
                  scan('embvar', '"#@@cvar"')
+    assert_equal ['#'],
+                 scan('embvar', '"#@1"')
     assert_equal [],
                  scan('embvar', '"#lvar"')
     assert_equal [],
@@ -264,6 +304,13 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('float', 'm(a,b,1.0,c,d)')
   end
 
+  def test_rational
+    assert_equal [],
+                 scan('rational', '')
+    assert_equal ['1r', '10r', '10.1r'],
+                 scan('rational', 'm(1r,10r,10.1r)')
+  end
+
   def test_gvar
     assert_equal [],
                  scan('gvar', '')
@@ -286,6 +333,13 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('ident', 'm(lvar, @ivar, @@cvar, $gvar)')
   end
 
+  def test_imaginary
+    assert_equal [],
+                 scan('imaginary', '')
+    assert_equal ['1i', '10ri', '10.0i', '10.1ri'],
+                 scan('imaginary', 'm(1i,10ri,10.0i,10.1ri)')
+  end
+
   def test_int
     assert_equal [],
                  scan('int', '')
@@ -480,8 +534,8 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('op', '1 < 1')
     assert_equal ['<='],
                  scan('op', '1 <= 1')
-    assert_equal ['<<'],
-                 scan('op', '1 << 1')
+    assert_equal ['<''<'],
+                 scan('op', '1 <''< 1')
     assert_equal ['>>'],
                  scan('op', '1 >> 1')
     assert_equal ['+'],
@@ -506,6 +560,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('op', ':[]')
     assert_equal ['[]='],
                  scan('op', ':[]=')
+    assert_equal ['&.'],
+                 scan('op', 'a&.f')
+    assert_equal %w(.:),
+                 scan('op', 'obj.:foo')
     assert_equal [],
                  scan('op', %q[`make all`])
   end
@@ -593,8 +651,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('words_beg', '%W()')
     assert_equal ['%W('],
                  scan('words_beg', '%W(w w w)')
-    assert_equal ['%W( '],
+    assert_equal ['%W('],
                  scan('words_beg', '%W( w w w )')
+    assert_equal ['%W('],
+                 scan('words_beg', "%W(\nw)")
   end
 
   def test_qwords_beg
@@ -604,8 +664,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('qwords_beg', '%w()')
     assert_equal ['%w('],
                  scan('qwords_beg', '%w(w w w)')
-    assert_equal ['%w( '],
+    assert_equal ['%w('],
                  scan('qwords_beg', '%w( w w w )')
+    assert_equal ['%w('],
+                 scan('qwords_beg', "%w(\nw)")
   end
 
   def test_qsymbols_beg
@@ -615,8 +677,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('qsymbols_beg', '%i()')
     assert_equal ['%i('],
                  scan('qsymbols_beg', '%i(w w w)')
-    assert_equal ['%i( '],
+    assert_equal ['%i('],
                  scan('qsymbols_beg', '%i( w w w )')
+    assert_equal ['%i('],
+                 scan('qsymbols_beg', "%i(\nw)")
   end
 
   def test_symbols_beg
@@ -626,72 +690,87 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('symbols_beg', '%I()')
     assert_equal ['%I('],
                  scan('symbols_beg', '%I(w w w)')
-    assert_equal ['%I( '],
+    assert_equal ['%I('],
                  scan('symbols_beg', '%I( w w w )')
+    assert_equal ['%I('],
+                 scan('symbols_beg', "%I(\nw)")
   end
 
-  # FIXME: Close paren must not present (`words_end' scanner event?).
   def test_words_sep
     assert_equal [],
                  scan('words_sep', '')
-    assert_equal [')'],
+    assert_equal [],
                  scan('words_sep', '%w()')
-    assert_equal [' ', ' ', ')'],
+    assert_equal [' ', ' '],
                  scan('words_sep', '%w(w w w)')
-    assert_equal [' ', ' ', ' )'],
+    assert_equal [' ', ' ', ' ', ' '],
                  scan('words_sep', '%w( w w w )')
-    assert_equal ["\n", ' ', ' )'],
+    assert_equal [' ', "\n", ' ', ' '],
                  scan('words_sep', "%w( w\nw w )")
+    assert_equal ["\n\n", "\n ", ' ', ' '],
+                 scan('words_sep', "%w(\n\nw\n w w )")
   end
 
   def test_heredoc_beg
     assert_equal [],
                  scan('heredoc_beg', '')
-    assert_equal ['<<EOS'],
-                 scan('heredoc_beg', "<<EOS\nheredoc\nEOS")
-    assert_equal ['<<EOS'],
-                 scan('heredoc_beg', "<<EOS\nheredoc\nEOS\n")
-    assert_equal ['<<EOS'],
-                 scan('heredoc_beg', "<<EOS\nheredoc\nEOS \n")
-    assert_equal ['<<-EOS'],
-                 scan('heredoc_beg', "<<-EOS\nheredoc\n\tEOS \n")
-    assert_equal ['<<"EOS"'],
-                 scan('heredoc_beg', '<<"EOS"'"\nheredoc\nEOS")
-    assert_equal ["<<'EOS'"],
-                 scan('heredoc_beg', "<<'EOS'\nheredoc\nEOS")
-    assert_equal ['<<`EOS`'],
-                 scan('heredoc_beg', "<<`EOS`\nheredoc\nEOS")
-    assert_equal ['<<" "'],
-                 scan('heredoc_beg', '<<" "'"\nheredoc\nEOS")
+    assert_equal ['<<''EOS'],
+                 scan('heredoc_beg', "<<""EOS\nheredoc\nEOS")
+    assert_equal ['<<''EOS'],
+                 scan('heredoc_beg', "<<""EOS\nheredoc\nEOS\n")
+    assert_equal ['<<''EOS'],
+                 scan('heredoc_beg', "<<""EOS\nheredoc\nEOS \n")
+    assert_equal ['<<''-EOS'],
+                 scan('heredoc_beg', "<<""-EOS\nheredoc\n\tEOS \n")
+    assert_equal ['<<''"EOS"'],
+                 scan('heredoc_beg', '<<''"EOS"'"\nheredoc\nEOS")
+    assert_equal ["<<""'EOS'"],
+                 scan('heredoc_beg', "<<""'EOS'\nheredoc\nEOS")
+    assert_equal ['<<''`EOS`'],
+                 scan('heredoc_beg', "<<""`EOS`\nheredoc\nEOS")
+    assert_equal ['<<''" "'],
+                 scan('heredoc_beg', '<<''" "'"\nheredoc\nEOS")
   end
 
   def test_tstring_content_HEREDOC
     assert_equal [],
                  scan('tstring_content', '')
     assert_equal ["heredoc\n"],
-                 scan('tstring_content', "<<EOS\nheredoc\nEOS")
+                 scan('tstring_content', "<<""EOS\nheredoc\nEOS")
     assert_equal ["heredoc\n"],
-                 scan('tstring_content', "<<EOS\nheredoc\nEOS\n")
+                 scan('tstring_content', "<<""EOS\nheredoc\nEOS\n")
     assert_equal ["here\ndoc \nEOS \n"],
-                 scan('tstring_content', "<<EOS\nhere\ndoc \nEOS \n")
+                 scan('tstring_content', "<<""EOS\nhere\ndoc \nEOS \n")
     assert_equal ["heredoc\n\tEOS \n"],
-                 scan('tstring_content', "<<-EOS\nheredoc\n\tEOS \n")
+                 scan('tstring_content', "<<""-EOS\nheredoc\n\tEOS \n")
+    bug7255 = '[ruby-core:48703]'
+    assert_equal ["there\n""heredoc", "\n"],
+                 scan('tstring_content', "<<""EOS\n""there\n""heredoc\#{foo}\nEOS"),
+                 bug7255
+    assert_equal ["there\n""heredoc", "\n"],
+                 scan('tstring_content', "<<""EOS\n""there\n""heredoc\#@foo\nEOS"),
+                 bug7255
+    bug10392 = '[ruby-dev:48647] [Bug #10392]'
+    assert_equal [" E\n\n"],
+                 scan('tstring_content', "<<""'E'\n E\n\n"),
+                 bug10392
+                 scan('tstring_content', "tap{<<""EOS}\n""there\n""heredoc\#@1xxx\nEOS")
   end
 
   def test_heredoc_end
     assert_equal [],
                  scan('heredoc_end', '')
     assert_equal ["EOS"],
-                 scan('heredoc_end', "<<EOS\nEOS"),
+                 scan('heredoc_end', "<<""EOS\nEOS"),
                  "bug#4543"
     assert_equal ["EOS"],
-                 scan('heredoc_end', "<<EOS\nheredoc\nEOS")
+                 scan('heredoc_end', "<<""EOS\nheredoc\nEOS")
     assert_equal ["EOS\n"],
-                 scan('heredoc_end', "<<EOS\nheredoc\nEOS\n")
+                 scan('heredoc_end', "<<""EOS\nheredoc\nEOS\n")
     assert_equal [],
-                 scan('heredoc_end', "<<EOS\nheredoc\nEOS \n")
+                 scan('heredoc_end', "<<""EOS\nheredoc\nEOS \n")
     assert_equal [],
-                 scan('heredoc_end', "<<-EOS\nheredoc\n\tEOS \n")
+                 scan('heredoc_end', "<<""-EOS\nheredoc\n\tEOS \n")
   end
 
   def test_semicolon
@@ -776,6 +855,11 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('sp', "%w(  w  )")
     assert_equal [],
                  scan('sp', "p(/ /)")
+
+    assert_equal ["\\\n"],
+                 scan('sp', "\\\n")
+    assert_equal ['\ '],
+                 scan('sp', '\ ')
   end
 
   # `nl' event always means End-Of-Statement.
@@ -817,6 +901,13 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('ignored_nl', "1;\r\n")
   end
 
+  def test_ignored_sp
+    assert_equal [],
+                 scan('ignored_sp', "<<~EOS\nheredoc\nEOS")
+    assert_equal ["  "],
+                 scan('ignored_sp', "<<~EOS\n  heredoc\nEOS")
+  end
+
   def test___end__
     assert_equal [],
                  scan('__end__', "")
@@ -835,22 +926,68 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
   def test_CHAR
     assert_equal [],
                  scan('CHAR', "")
-    assert_equal ["@"],
-                 scan('CHAR', "@")
+    assert_equal ["?a"],
+                 scan('CHAR', "?a")
     assert_equal [],
                  scan('CHAR', "@ivar")
+
+    assert_equal ["?\\M-H"], scan('CHAR', '?\\M-H')
+
+    assert_equal ["?\\u0041"],
+                 scan('CHAR', "?\\u0041")
+
+    assert_equal ["?\\u{41}"],
+                 scan('CHAR', "?\\u{41}")
+
+    err = nil
+    assert_equal [], scan('CHAR', '?\\M ') {|*e| err = e}
+    assert_equal([:on_parse_error, "Invalid escape character syntax", "?\\M "], err)
+
+    err = nil
+    scan('CHAR', '?\u{41 42}') {|*e| err = e}
+    assert_equal [:on_parse_error, "Multiple codepoints at single character literal", "42"], err
   end
 
   def test_label
+    assert_equal %w(foo:),
+                 scan('label', '{foo: 1}')
+  end
+
+  def test_label_end
+    assert_equal %w(":),
+                 scan('label_end', '{"foo-bar": 1}')
   end
 
   def test_tlambda
+    assert_equal %w(->),
+                 scan('tlambda', '->{}')
   end
 
   def test_tlambeg
+    assert_equal %w({),
+                 scan('tlambeg', '-> {}')
   end
 
   def test_tlambda_arg
+    assert_equal %w(),
+                 scan('tlambda_arg', '-> {}')
   end
 
+  def test_invalid_char
+    err = nil
+    assert_equal(['a'], scan('ident', "\ea") {|*e| err = e})
+    assert_equal(:compile_error, err[0])
+    assert_match(/Invalid char/, err[1])
+    assert_equal("\e", err[2])
+  end
+
+  def test_invalid_hex_escape
+    err = nil
+    assert_equal ['U'], scan('tstring_content', '"\\xU"') {|*e| err = e}
+    assert_equal [:on_parse_error, "invalid hex escape", "\\x"], err
+
+    err = nil
+    assert_equal ['U'], scan('tstring_content', '/\\xU/') {|*e| err = e}
+    assert_equal [:on_parse_error, "invalid hex escape", "\\x"], err
+  end
 end if ripper_test
