@@ -1,12 +1,14 @@
+# frozen_string_literal: false
 require "test/unit"
 require "fileutils"
 require "tmpdir"
 require "socket"
+require '-test-/file'
 
 class TestFileExhaustive < Test::Unit::TestCase
   DRIVE = Dir.pwd[%r'\A(?:[a-z]:|//[^/]+/[^/]+)'i]
   POSIX = /cygwin|mswin|bccwin|mingw|emx/ !~ RUBY_PLATFORM
-  NTFS = !(/cygwin|mingw|mswin|bccwin/ !~ RUBY_PLATFORM)
+  NTFS = !(/mingw|mswin|bccwin/ !~ RUBY_PLATFORM)
 
   def assert_incompatible_encoding
     d = "\u{3042}\u{3044}".encode("utf-16le")
@@ -59,9 +61,16 @@ class TestFileExhaustive < Test::Unit::TestCase
     @file
   end
 
+  def utf8_file
+    return @utf8file if defined? @utf8file
+    @utf8file = make_tmp_filename("\u3066\u3059\u3068")
+    make_file("foo", @utf8file)
+    @utf8file
+  end
+
   def notownedfile
     return @notownedfile if defined? @notownedfile
-    if Process.euid != 0
+    if Process.euid != File.stat("/").uid
       @notownedfile = '/'
     else
       @notownedfile = nil
@@ -106,11 +115,11 @@ class TestFileExhaustive < Test::Unit::TestCase
   end
 
   def symlinkfile
-    return @symlinkfile if @symlinkfile
+    return @symlinkfile if defined? @symlinkfile
     @symlinkfile = make_tmp_filename("symlinkfile")
     begin
       File.symlink(regular_file, @symlinkfile)
-    rescue NotImplementedError, Errno::EACCES
+    rescue NotImplementedError, Errno::EACCES, Errno::EPERM
       @symlinkfile = nil
     end
     @symlinkfile
@@ -167,15 +176,33 @@ class TestFileExhaustive < Test::Unit::TestCase
   end
 
   def test_path
-    file = regular_file
+    [regular_file, utf8_file].each do |file|
+      assert_equal(file, File.open(file) {|f| f.path})
+      assert_equal(file, File.path(file))
+      o = Object.new
+      class << o; self; end.class_eval do
+        define_method(:to_path) { file }
+      end
+      assert_equal(file, File.path(o))
+    end
+  end
 
-    assert_equal(file, File.open(file) {|f| f.path})
-    assert_equal(file, File.path(file))
-    o = Object.new
-    class << o; self; end.class_eval do
-      define_method(:to_path) { file }
+  def test_path_taint
+    [regular_file, utf8_file].each do |file|
+      file.untaint
+      assert_equal(false, File.open(file) {|f| f.path}.tainted?)
+      assert_equal(true, File.open(file.dup.taint) {|f| f.path}.tainted?)
+      o = Object.new
+      class << o; self; end.class_eval do
+        define_method(:to_path) { file }
+      end
+      assert_equal(false, File.open(o) {|f| f.path}.tainted?)
+      class << o; self; end.class_eval do
+        remove_method(:to_path)
+        define_method(:to_path) { file.dup.taint }
+      end
+      assert_equal(true, File.open(o) {|f| f.path}.tainted?)
     end
-    assert_equal(file, File.path(o))
   end
 
   def assert_integer(n)
@@ -183,7 +210,7 @@ class TestFileExhaustive < Test::Unit::TestCase
   end
 
   def assert_integer_or_nil(n)
-    msg = ->{"#{n.inspect} is neither Fixnum nor nil."}
+    msg = ->{"#{n.inspect} is neither Integer nor nil."}
     if n
       assert_kind_of(Integer, n, msg)
     else
@@ -254,16 +281,28 @@ class TestFileExhaustive < Test::Unit::TestCase
     end
   end if NTFS
 
+  def test_lstat
+    return unless symlinkfile
+    assert_equal(false, File.stat(symlinkfile).symlink?)
+    assert_equal(true, File.lstat(symlinkfile).symlink?)
+    f = File.new(symlinkfile)
+    assert_equal(false, f.stat.symlink?)
+    assert_equal(true, f.lstat.symlink?)
+    f.close
+  end
+
   def test_directory_p
     assert_file.directory?(@dir)
     assert_file.not_directory?(@dir+"/...")
     assert_file.not_directory?(regular_file)
+    assert_file.not_directory?(utf8_file)
     assert_file.not_directory?(nofile)
   end
 
   def test_pipe_p
     assert_file.not_pipe?(@dir)
     assert_file.not_pipe?(regular_file)
+    assert_file.not_pipe?(utf8_file)
     assert_file.not_pipe?(nofile)
     assert_file.pipe?(fifo) if fifo
   end
@@ -271,6 +310,7 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_symlink_p
     assert_file.not_symlink?(@dir)
     assert_file.not_symlink?(regular_file)
+    assert_file.not_symlink?(utf8_file)
     assert_file.symlink?(symlinkfile) if symlinkfile
     assert_file.not_symlink?(hardlinkfile) if hardlinkfile
     assert_file.not_symlink?(nofile)
@@ -279,6 +319,7 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_socket_p
     assert_file.not_socket?(@dir)
     assert_file.not_socket?(regular_file)
+    assert_file.not_socket?(utf8_file)
     assert_file.not_socket?(nofile)
     assert_file.socket?(socket) if socket
   end
@@ -286,6 +327,7 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_blockdev_p
     assert_file.not_blockdev?(@dir)
     assert_file.not_blockdev?(regular_file)
+    assert_file.not_blockdev?(utf8_file)
     assert_file.not_blockdev?(nofile)
     assert_file.blockdev?(blockdev) if blockdev
   end
@@ -293,6 +335,7 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_chardev_p
     assert_file.not_chardev?(@dir)
     assert_file.not_chardev?(regular_file)
+    assert_file.not_chardev?(utf8_file)
     assert_file.not_chardev?(nofile)
     assert_file.chardev?(chardev) if chardev
   end
@@ -300,6 +343,7 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_exist_p
     assert_file.exist?(@dir)
     assert_file.exist?(regular_file)
+    assert_file.exist?(utf8_file)
     assert_file.not_exist?(nofile)
   end
 
@@ -309,6 +353,12 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_readable?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.readable?(regular_file)
+
+    File.chmod(0200, utf8_file)
+    assert_file.not_readable?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.readable?(utf8_file)
+
     assert_file.not_readable?(nofile)
   end if POSIX
 
@@ -318,6 +368,12 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_readable_real?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.readable_real?(regular_file)
+
+    File.chmod(0200, utf8_file)
+    assert_file.not_readable_real?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.readable_real?(utf8_file)
+
     assert_file.not_readable_real?(nofile)
   end if POSIX
 
@@ -328,6 +384,14 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_world_readable?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.not_world_readable?(regular_file)
+
+    File.chmod(0006, utf8_file)
+    assert_file.world_readable?(utf8_file)
+    File.chmod(0060, utf8_file)
+    assert_file.not_world_readable?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.not_world_readable?(utf8_file)
+
     assert_file.not_world_readable?(nofile)
   end if POSIX
 
@@ -337,6 +401,12 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_writable?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.writable?(regular_file)
+
+    File.chmod(0400, utf8_file)
+    assert_file.not_writable?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.writable?(utf8_file)
+
     assert_file.not_writable?(nofile)
   end if POSIX
 
@@ -346,6 +416,12 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_writable_real?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.writable_real?(regular_file)
+
+    File.chmod(0400, utf8_file)
+    assert_file.not_writable_real?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.writable_real?(utf8_file)
+
     assert_file.not_writable_real?(nofile)
   end if POSIX
 
@@ -356,6 +432,14 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_world_writable?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.not_world_writable?(regular_file)
+
+    File.chmod(0006, utf8_file)
+    assert_file.world_writable?(utf8_file)
+    File.chmod(0060, utf8_file)
+    assert_file.not_world_writable?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.not_world_writable?(utf8_file)
+
     assert_file.not_world_writable?(nofile)
   end if POSIX
 
@@ -364,6 +448,12 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.executable?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.not_executable?(regular_file)
+
+    File.chmod(0100, utf8_file)
+    assert_file.executable?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.not_executable?(utf8_file)
+
     assert_file.not_executable?(nofile)
   end if POSIX
 
@@ -372,51 +462,90 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.executable_real?(regular_file)
     File.chmod(0600, regular_file)
     assert_file.not_executable_real?(regular_file)
+
+    File.chmod(0100, utf8_file)
+    assert_file.executable_real?(utf8_file)
+    File.chmod(0600, utf8_file)
+    assert_file.not_executable_real?(utf8_file)
+
     assert_file.not_executable_real?(nofile)
   end if POSIX
 
   def test_file_p
     assert_file.not_file?(@dir)
     assert_file.file?(regular_file)
+    assert_file.file?(utf8_file)
     assert_file.not_file?(nofile)
   end
 
   def test_zero_p
     assert_nothing_raised { File.zero?(@dir) }
     assert_file.not_zero?(regular_file)
+    assert_file.not_zero?(utf8_file)
     assert_file.zero?(zerofile)
     assert_file.not_zero?(nofile)
   end
 
+  def test_empty_p
+    assert_nothing_raised { File.empty?(@dir) }
+    assert_file.not_empty?(regular_file)
+    assert_file.not_empty?(utf8_file)
+    assert_file.empty?(zerofile)
+    assert_file.not_empty?(nofile)
+  end
+
   def test_size_p
     assert_nothing_raised { File.size?(@dir) }
     assert_equal(3, File.size?(regular_file))
+    assert_equal(3, File.size?(utf8_file))
     assert_file.not_size?(zerofile)
     assert_file.not_size?(nofile)
   end
 
   def test_owned_p
     assert_file.owned?(regular_file)
+    assert_file.owned?(utf8_file)
     assert_file.not_owned?(notownedfile) if notownedfile
   end if POSIX
 
   def test_grpowned_p ## xxx
     assert_file.grpowned?(regular_file)
+    assert_file.grpowned?(utf8_file)
   end if POSIX
 
+  def io_open(file_name)
+    # avoid File.open since we do not want #to_path
+    io = IO.for_fd(IO.sysopen(file_name))
+    yield io
+  ensure
+    io&.close
+  end
+
   def test_suid
     assert_file.not_setuid?(regular_file)
-    assert_file.setuid?(suidfile) if suidfile
+    assert_file.not_setuid?(utf8_file)
+    if suidfile
+      assert_file.setuid?(suidfile)
+      io_open(suidfile) { |io| assert_file.setuid?(io) }
+    end
   end
 
   def test_sgid
     assert_file.not_setgid?(regular_file)
-    assert_file.setgid?(sgidfile) if sgidfile
+    assert_file.not_setgid?(utf8_file)
+    if sgidfile
+      assert_file.setgid?(sgidfile)
+      io_open(sgidfile) { |io| assert_file.setgid?(io) }
+    end
   end
 
   def test_sticky
     assert_file.not_sticky?(regular_file)
-    assert_file.sticky?(stickyfile) if stickyfile
+    assert_file.not_sticky?(utf8_file)
+    if stickyfile
+      assert_file.sticky?(stickyfile)
+      io_open(stickyfile) { |io| assert_file.sticky?(io) }
+    end
   end
 
   def test_path_identical_p
@@ -426,26 +555,40 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_file.not_identical?(nofile, regular_file)
   end
 
+  def path_identical_p(file)
+    [regular_file, utf8_file].each do |file|
+      assert_file.identical?(file, file)
+      assert_file.not_identical?(file, zerofile)
+      assert_file.not_identical?(file, nofile)
+      assert_file.not_identical?(nofile, file)
+    end
+  end
+
   def test_io_identical_p
-    open(regular_file) {|f|
-      assert_file.identical?(f, f)
-      assert_file.identical?(regular_file, f)
-      assert_file.identical?(f, regular_file)
-    }
+    [regular_file, utf8_file].each do |file|
+      open(file) {|f|
+        assert_file.identical?(f, f)
+        assert_file.identical?(file, f)
+        assert_file.identical?(f, file)
+      }
+    end
   end
 
   def test_closed_io_identical_p
-    io = open(regular_file) {|f| f}
-    assert_raise(IOError) {
-      File.identical?(regular_file, io)
-    }
-    File.unlink(regular_file)
-    assert_file.not_exist?(regular_file)
+    [regular_file, utf8_file].each do |file|
+      io = open(file) {|f| f}
+      assert_raise(IOError) {
+        File.identical?(file, io)
+      }
+      File.unlink(file)
+      assert_file.not_exist?(file)
+    end
   end
 
   def test_s_size
     assert_integer(File.size(@dir))
     assert_equal(3, File.size(regular_file))
+    assert_equal(3, File.size(utf8_file))
     assert_equal(0, File.size(zerofile))
     assert_raise(Errno::ENOENT) { File.size(nofile) }
   end
@@ -453,51 +596,85 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_ftype
     assert_equal("directory", File.ftype(@dir))
     assert_equal("file", File.ftype(regular_file))
+    assert_equal("file", File.ftype(utf8_file))
     assert_equal("link", File.ftype(symlinkfile)) if symlinkfile
     assert_equal("file", File.ftype(hardlinkfile)) if hardlinkfile
     assert_raise(Errno::ENOENT) { File.ftype(nofile) }
   end
 
   def test_atime
-    t1 = File.atime(regular_file)
-    t2 = File.open(regular_file) {|f| f.atime}
-    assert_kind_of(Time, t1)
-    assert_kind_of(Time, t2)
-    assert_equal(t1, t2)
+    [regular_file, utf8_file].each do |file|
+      t1 = File.atime(file)
+      t2 = File.open(file) {|f| f.atime}
+      assert_kind_of(Time, t1)
+      assert_kind_of(Time, t2)
+      # High Sierra's APFS can handle nano-sec precise.
+      # t1 value is difference from t2 on APFS.
+      if Bug::File::Fs.fsname(Dir.tmpdir) == "apfs"
+        assert_equal(t1.to_i, t2.to_i)
+      else
+        assert_equal(t1, t2)
+      end
+    end
     assert_raise(Errno::ENOENT) { File.atime(nofile) }
   end
 
   def test_mtime
-    t1 = File.mtime(regular_file)
-    t2 = File.open(regular_file) {|f| f.mtime}
-    assert_kind_of(Time, t1)
-    assert_kind_of(Time, t2)
-    assert_equal(t1, t2)
+    [regular_file, utf8_file].each do |file|
+      t1 = File.mtime(file)
+      t2 = File.open(file) {|f| f.mtime}
+      assert_kind_of(Time, t1)
+      assert_kind_of(Time, t2)
+      assert_equal(t1, t2)
+    end
     assert_raise(Errno::ENOENT) { File.mtime(nofile) }
   end
 
   def test_ctime
-    t1 = File.ctime(regular_file)
-    t2 = File.open(regular_file) {|f| f.ctime}
-    assert_kind_of(Time, t1)
-    assert_kind_of(Time, t2)
-    assert_equal(t1, t2)
+    [regular_file, utf8_file].each do |file|
+      t1 = File.ctime(file)
+      t2 = File.open(file) {|f| f.ctime}
+      assert_kind_of(Time, t1)
+      assert_kind_of(Time, t2)
+      assert_equal(t1, t2)
+    end
     assert_raise(Errno::ENOENT) { File.ctime(nofile) }
   end
 
+  def test_birthtime
+    skip if RUBY_PLATFORM =~ /android/
+    [regular_file, utf8_file].each do |file|
+      t1 = File.birthtime(file)
+      t2 = File.open(file) {|f| f.birthtime}
+      assert_kind_of(Time, t1)
+      assert_kind_of(Time, t2)
+      assert_equal(t1, t2)
+    rescue Errno::ENOSYS, NotImplementedError
+      # ignore unsupporting filesystems
+    rescue Errno::EPERM
+      # Docker prohibits statx syscall by the default.
+      skip("statx(2) is prohibited by seccomp")
+    end
+    assert_raise(Errno::ENOENT) { File.birthtime(nofile) }
+  end if File.respond_to?(:birthtime)
+
   def test_chmod
-    assert_equal(1, File.chmod(0444, regular_file))
-    assert_equal(0444, File.stat(regular_file).mode % 01000)
-    assert_equal(0, File.open(regular_file) {|f| f.chmod(0222)})
-    assert_equal(0222, File.stat(regular_file).mode % 01000)
-    File.chmod(0600, regular_file)
+    [regular_file, utf8_file].each do |file|
+      assert_equal(1, File.chmod(0444, file))
+      assert_equal(0444, File.stat(file).mode % 01000)
+      assert_equal(0, File.open(file) {|f| f.chmod(0222)})
+      assert_equal(0222, File.stat(file).mode % 01000)
+      File.chmod(0600, file)
+    end
     assert_raise(Errno::ENOENT) { File.chmod(0600, nofile) }
   end if POSIX
 
   def test_lchmod
-    assert_equal(1, File.lchmod(0444, regular_file))
-    assert_equal(0444, File.stat(regular_file).mode % 01000)
-    File.lchmod(0600, regular_file)
+    [regular_file, utf8_file].each do |file|
+      assert_equal(1, File.lchmod(0444, file))
+      assert_equal(0444, File.stat(file).mode % 01000)
+      File.lchmod(0600, regular_file)
+    end
     assert_raise(Errno::ENOENT) { File.lchmod(0600, nofile) }
   rescue NotImplementedError
   end if POSIX
@@ -512,6 +689,7 @@ class TestFileExhaustive < Test::Unit::TestCase
     return unless symlinkfile
     assert_equal("link", File.ftype(symlinkfile))
     assert_raise(Errno::EEXIST) { File.symlink(regular_file, regular_file) }
+    assert_raise(Errno::EEXIST) { File.symlink(utf8_file, utf8_file) }
   end
 
   def test_utime
@@ -521,16 +699,45 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_equal(t + 2, File.mtime(zerofile))
   end
 
+  def test_utime_symlinkfile
+    return unless symlinkfile
+    t = Time.local(2000)
+    assert_equal(1, File.utime(t, t, symlinkfile))
+    assert_equal(t, File.stat(regular_file).atime)
+    assert_equal(t, File.stat(regular_file).mtime)
+  end
+
+  def test_lutime
+    return unless File.respond_to?(:lutime)
+    return unless symlinkfile
+
+    r = File.stat(regular_file)
+    t = Time.local(2000)
+    File.lutime(t + 1, t + 2, symlinkfile)
+  rescue NotImplementedError => e
+    skip(e.message)
+  else
+    stat = File.stat(regular_file)
+    assert_equal(r.atime, stat.atime)
+    assert_equal(r.mtime, stat.mtime)
+
+    stat = File.lstat(symlinkfile)
+    assert_equal(t + 1, stat.atime)
+    assert_equal(t + 2, stat.mtime)
+  end
+
   def test_hardlink
     return unless hardlinkfile
     assert_equal("file", File.ftype(hardlinkfile))
     assert_raise(Errno::EEXIST) { File.link(regular_file, regular_file) }
+    assert_raise(Errno::EEXIST) { File.link(utf8_file, utf8_file) }
   end
 
   def test_readlink
     return unless symlinkfile
     assert_equal(regular_file, File.readlink(symlinkfile))
     assert_raise(Errno::EINVAL) { File.readlink(regular_file) }
+    assert_raise(Errno::EINVAL) { File.readlink(utf8_file) }
     assert_raise(Errno::ENOENT) { File.readlink(nofile) }
     if fs = Encoding.find("filesystem")
       assert_equal(fs, File.readlink(symlinkfile).encoding)
@@ -541,7 +748,8 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_readlink_long_path
     return unless symlinkfile
     bug9157 = '[ruby-core:58592] [Bug #9157]'
-    assert_separately(["-", symlinkfile, bug9157], <<-"end;")
+    assert_separately(["-", symlinkfile, bug9157], "#{<<~begin}#{<<~"end;"}")
+    begin
       symlinkfile, bug9157 = *ARGV
       100.step(1000, 100) do |n|
         File.unlink(symlinkfile)
@@ -556,18 +764,42 @@ class TestFileExhaustive < Test::Unit::TestCase
     end;
   end
 
+  if NTFS
+    def test_readlink_junction
+      base = File.basename(nofile)
+      err = IO.popen(%W"cmd.exe /c mklink /j #{base} .", chdir: @dir, err: %i[child out], &:read)
+      skip err unless $?.success?
+      assert_equal(@dir, File.readlink(nofile))
+    end
+
+    def test_realpath_mount_point
+      vol = IO.popen(["mountvol", DRIVE, "/l"], &:read).strip
+      Dir.mkdir(mnt = File.join(@dir, mntpnt = "mntpnt"))
+      system("mountvol", mntpnt, vol, chdir: @dir)
+      assert_equal(mnt, File.realpath(mnt))
+    ensure
+      system("mountvol", mntpnt, "/d", chdir: @dir)
+    end
+  end
+
   def test_unlink
     assert_equal(1, File.unlink(regular_file))
     make_file("foo", regular_file)
+
+    assert_equal(1, File.unlink(utf8_file))
+    make_file("foo", utf8_file)
+
     assert_raise(Errno::ENOENT) { File.unlink(nofile) }
   end
 
   def test_rename
-    assert_equal(0, File.rename(regular_file, nofile))
-    assert_file.not_exist?(regular_file)
-    assert_file.exist?(nofile)
-    assert_equal(0, File.rename(nofile, regular_file))
-    assert_raise(Errno::ENOENT) { File.rename(nofile, regular_file) }
+    [regular_file, utf8_file].each do |file|
+      assert_equal(0, File.rename(file, nofile))
+      assert_file.not_exist?(file)
+      assert_file.exist?(nofile)
+      assert_equal(0, File.rename(nofile, file))
+      assert_raise(Errno::ENOENT) { File.rename(nofile, file) }
+    end
   end
 
   def test_umask
@@ -582,33 +814,51 @@ class TestFileExhaustive < Test::Unit::TestCase
 
   def test_expand_path
     assert_equal(regular_file, File.expand_path(File.basename(regular_file), File.dirname(regular_file)))
-    if NTFS
-      assert_equal(regular_file, File.expand_path(regular_file + " "))
-      assert_equal(regular_file, File.expand_path(regular_file + "."))
-      assert_equal(regular_file, File.expand_path(regular_file + "::$DATA"))
+    assert_equal(utf8_file, File.expand_path(File.basename(utf8_file), File.dirname(utf8_file)))
+  end
+
+  if NTFS
+    def test_expand_path_ntfs
+      [regular_file, utf8_file].each do |file|
+        assert_equal(file, File.expand_path(file + " "))
+        assert_equal(file, File.expand_path(file + "."))
+        assert_equal(file, File.expand_path(file + "::$DATA"))
+      end
       assert_match(/\Ac:\//i, File.expand_path('c:'), '[ruby-core:31591]')
       assert_match(/\Ac:\//i, File.expand_path('c:foo', 'd:/bar'))
       assert_match(/\Ae:\//i, File.expand_path('e:foo', 'd:/bar'))
       assert_match(%r'\Ac:/bar/foo\z'i, File.expand_path('c:foo', 'c:/bar'))
     end
-    case RUBY_PLATFORM
-    when /darwin/
+  end
+
+  case RUBY_PLATFORM
+  when /darwin/
+    def test_expand_path_hfs
       ["\u{feff}", *"\u{2000}"..."\u{2100}"].each do |c|
         file = regular_file + c
+        full_path = File.expand_path(file)
+        mesg = proc {File.basename(full_path).dump}
         begin
           open(file) {}
         rescue
-          assert_equal(file, File.expand_path(file), c.dump)
+          # High Sierra's APFS cannot use filenames with undefined character
+          next if Bug::File::Fs.fsname(Dir.tmpdir) == "apfs"
+          assert_equal(file, full_path, mesg)
         else
-          assert_equal(regular_file, File.expand_path(file), c.dump)
+          assert_equal(regular_file, full_path, mesg)
         end
       end
     end
-    if DRIVE
+  end
+
+  if DRIVE
+    def test_expand_path_absolute
       assert_match(%r"\Az:/foo\z"i, File.expand_path('/foo', "z:/bar"))
       assert_match(%r"\A//host/share/foo\z"i, File.expand_path('/foo', "//host/share/bar"))
       assert_match(%r"\A#{DRIVE}/foo\z"i, File.expand_path('/foo'))
-    else
+    end
+  else
+    def test_expand_path_absolute
       assert_equal("/foo", File.expand_path('/foo'))
     end
   end
@@ -633,6 +883,8 @@ class TestFileExhaustive < Test::Unit::TestCase
     a = "#{drive}/\225\\\\"
     if File::ALT_SEPARATOR == '\\'
       [%W"cp437 #{drive}/\225", %W"cp932 #{drive}/\225\\"]
+    elsif File.directory?("#{@dir}/\\")
+      [%W"cp437 /\225", %W"cp932 /\225\\"]
     else
       [["cp437", a], ["cp932", a]]
     end.each do |cp, expected|
@@ -678,7 +930,6 @@ class TestFileExhaustive < Test::Unit::TestCase
       ENV["HOMEDRIVE"] = nil
       ENV["HOMEPATH"] = nil
       ENV["USERPROFILE"] = nil
-      assert_raise(ArgumentError) { File.expand_path("~") }
       ENV["HOME"] = "~"
       assert_raise(ArgumentError, bug3630) { File.expand_path("~") }
       ENV["HOME"] = "."
@@ -702,6 +953,8 @@ class TestFileExhaustive < Test::Unit::TestCase
 
     assert_raise(ArgumentError) { File.expand_path(".", UnknownUserHome) }
     assert_nothing_raised(ArgumentError) { File.expand_path("#{DRIVE}/", UnknownUserHome) }
+    ENV["HOME"] = "#{DRIVE}UserHome"
+    assert_raise(ArgumentError) { File.expand_path("~") }
   ensure
     ENV["HOME"] = home
   end
@@ -945,8 +1198,29 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_equal('z:/bar/foo', File.expand_path('z:foo', '/bar'), bug10858)
   end if DRIVE
 
+  if /darwin/ =~ RUBY_PLATFORM and Encoding.find("filesystem") == Encoding::UTF_8
+    def test_expand_path_compose
+      pp = Object.new.extend(Test::Unit::Assertions)
+      def pp.mu_pp(str) #:nodoc:
+        str.dump
+      end
+
+      Dir.mktmpdir do |dir|
+        Dir.chdir(dir) do
+          orig = %W"d\u{e9}tente x\u{304c 304e 3050 3052 3054}"
+          orig.each do |o|
+            Dir.mkdir(o)
+            n = Dir.chdir(o) {File.expand_path(".")}
+            pp.assert_equal(o, File.basename(n))
+          end
+        end
+      end
+    end
+  end
+
   def test_basename
     assert_equal(File.basename(regular_file).sub(/\.test$/, ""), File.basename(regular_file, ".test"))
+    assert_equal(File.basename(utf8_file).sub(/\.test$/, ""), File.basename(utf8_file, ".test"))
     assert_equal("", s = File.basename(""))
     assert_not_predicate(s, :frozen?, '[ruby-core:24199]')
     assert_equal("foo", s = File.basename("foo"))
@@ -954,26 +1228,52 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_equal("foo", File.basename("foo", ".ext"))
     assert_equal("foo", File.basename("foo.ext", ".ext"))
     assert_equal("foo", File.basename("foo.ext", ".*"))
-    if NTFS
-      basename = File.basename(regular_file)
-      assert_equal(basename, File.basename(regular_file + " "))
-      assert_equal(basename, File.basename(regular_file + "."))
-      assert_equal(basename, File.basename(regular_file + "::$DATA"))
-      basename.chomp!(".test")
-      assert_equal(basename, File.basename(regular_file + " ", ".test"))
-      assert_equal(basename, File.basename(regular_file + ".", ".test"))
-      assert_equal(basename, File.basename(regular_file + "::$DATA", ".test"))
-      assert_equal(basename, File.basename(regular_file + " ", ".*"))
-      assert_equal(basename, File.basename(regular_file + ".", ".*"))
-      assert_equal(basename, File.basename(regular_file + "::$DATA", ".*"))
+  end
+
+  if NTFS
+    def test_basename_strip
+      [regular_file, utf8_file].each do |file|
+        basename = File.basename(file)
+        assert_equal(basename, File.basename(file + " "))
+        assert_equal(basename, File.basename(file + "."))
+        assert_equal(basename, File.basename(file + "::$DATA"))
+        basename.chomp!(".test")
+        assert_equal(basename, File.basename(file + " ", ".test"))
+        assert_equal(basename, File.basename(file + ".", ".test"))
+        assert_equal(basename, File.basename(file + "::$DATA", ".test"))
+        assert_equal(basename, File.basename(file + " ", ".*"))
+        assert_equal(basename, File.basename(file + ".", ".*"))
+        assert_equal(basename, File.basename(file + "::$DATA", ".*"))
+      end
     end
-    if File::ALT_SEPARATOR == '\\'
+  else
+    def test_basename_strip
+      [regular_file, utf8_file].each do |file|
+        basename = File.basename(file)
+        assert_equal(basename + " ", File.basename(file + " "))
+        assert_equal(basename + ".", File.basename(file + "."))
+        assert_equal(basename + "::$DATA", File.basename(file + "::$DATA"))
+        assert_equal(basename + " ", File.basename(file + " ", ".test"))
+        assert_equal(basename + ".", File.basename(file + ".", ".test"))
+        assert_equal(basename + "::$DATA", File.basename(file + "::$DATA", ".test"))
+        assert_equal(basename, File.basename(file + ".", ".*"))
+        basename.chomp!(".test")
+        assert_equal(basename, File.basename(file + " ", ".*"))
+        assert_equal(basename, File.basename(file + "::$DATA", ".*"))
+      end
+    end
+  end
+
+  if File::ALT_SEPARATOR == '\\'
+    def test_basename_backslash
       a = "foo/\225\\\\"
       [%W"cp437 \225", %W"cp932 \225\\"].each do |cp, expected|
         assert_equal(expected.force_encoding(cp), File.basename(a.dup.force_encoding(cp)), cp)
       end
     end
+  end
 
+  def test_basename_encoding
     assert_incompatible_encoding {|d| File.basename(d)}
     assert_incompatible_encoding {|d| File.basename(d, ".*")}
     assert_raise(Encoding::CompatibilityError) {File.basename("foo.ext", ".*".encode("utf-16le"))}
@@ -986,10 +1286,17 @@ class TestFileExhaustive < Test::Unit::TestCase
   end
 
   def test_dirname
-    assert(regular_file.start_with?(File.dirname(regular_file)))
+    assert_equal(@dir, File.dirname(regular_file))
+    assert_equal(@dir, File.dirname(utf8_file))
     assert_equal(".", File.dirname(""))
+  end
+
+  def test_dirname_encoding
     assert_incompatible_encoding {|d| File.dirname(d)}
-    if File::ALT_SEPARATOR == '\\'
+  end
+
+  if File::ALT_SEPARATOR == '\\'
+    def test_dirname_backslash
       a = "\225\\\\foo"
       [%W"cp437 \225", %W"cp932 \225\\"].each do |cp, expected|
         assert_equal(expected.force_encoding(cp), File.dirname(a.dup.force_encoding(cp)), cp)
@@ -999,22 +1306,25 @@ class TestFileExhaustive < Test::Unit::TestCase
 
   def test_extname
     assert_equal(".test", File.extname(regular_file))
+    assert_equal(".test", File.extname(utf8_file))
     prefixes = ["", "/", ".", "/.", "bar/.", "/bar/."]
-    infixes = ["", " ", "."]
+    infixes = ["", " "]
     infixes2 = infixes + [".ext "]
     appendixes = [""]
     if NTFS
       appendixes << " " << "." << "::$DATA" << "::$DATA.bar"
+    else
+      appendixes << [".", "."]
     end
     prefixes.each do |prefix|
-      appendixes.each do |appendix|
+      appendixes.each do |appendix, ext = ""|
         infixes.each do |infix|
           path = "#{prefix}foo#{infix}#{appendix}"
-          assert_equal("", File.extname(path), "File.extname(#{path.inspect})")
+          assert_equal(ext, File.extname(path), "File.extname(#{path.inspect})")
         end
         infixes2.each do |infix|
           path = "#{prefix}foo#{infix}.ext#{appendix}"
-          assert_equal(".ext", File.extname(path), "File.extname(#{path.inspect})")
+          assert_equal(ext.empty? ? ".ext" : appendix, File.extname(path), "File.extname(#{path.inspect})")
         end
       end
     end
@@ -1025,9 +1335,11 @@ class TestFileExhaustive < Test::Unit::TestCase
   end
 
   def test_split
-    d, b = File.split(regular_file)
-    assert_equal(File.dirname(regular_file), d)
-    assert_equal(File.basename(regular_file), b)
+    [regular_file, utf8_file].each do |file|
+      d, b = File.split(file)
+      assert_equal(File.dirname(file), d)
+      assert_equal(File.basename(file), b)
+    end
   end
 
   def test_join
@@ -1069,36 +1381,94 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_raise(Encoding::CompatibilityError, bug7168) {File.join(names)}
   end
 
+  def test_join_with_changed_separator
+    assert_separately([], "#{<<~"begin;"}\n#{<<~"end;"}")
+    bug = '[ruby-core:79579] [Bug #13223]'
+    begin;
+      class File
+        remove_const :Separator
+        remove_const :SEPARATOR
+      end
+      GC.start
+      assert_equal("hello/world", File.join("hello", "world"), bug)
+    end;
+  end
+
   def test_truncate
-    assert_equal(0, File.truncate(regular_file, 1))
-    assert_file.exist?(regular_file)
-    assert_equal(1, File.size(regular_file))
-    assert_equal(0, File.truncate(regular_file, 0))
-    assert_file.exist?(regular_file)
-    assert_file.zero?(regular_file)
-    make_file("foo", regular_file)
-    assert_raise(Errno::ENOENT) { File.truncate(nofile, 0) }
+    [regular_file, utf8_file].each do |file|
+      assert_equal(0, File.truncate(file, 1))
+      assert_file.exist?(file)
+      assert_equal(1, File.size(file))
+      assert_equal(0, File.truncate(file, 0))
+      assert_file.exist?(file)
+      assert_file.zero?(file)
+      make_file("foo", file)
+      assert_raise(Errno::ENOENT) { File.truncate(nofile, 0) }
 
-    f = File.new(regular_file, "w")
-    assert_equal(0, f.truncate(2))
-    assert_file.exist?(regular_file)
-    assert_equal(2, File.size(regular_file))
-    assert_equal(0, f.truncate(0))
-    assert_file.exist?(regular_file)
-    assert_file.zero?(regular_file)
-    f.close
-    make_file("foo", regular_file)
+      f = File.new(file, "w")
+      assert_equal(0, f.truncate(2))
+      assert_file.exist?(file)
+      assert_equal(2, File.size(file))
+      assert_equal(0, f.truncate(0))
+      assert_file.exist?(file)
+      assert_file.zero?(file)
+      f.close
+      make_file("foo", file)
 
-    assert_raise(IOError) { File.open(regular_file) {|ff| ff.truncate(0)} }
+      assert_raise(IOError) { File.open(file) {|ff| ff.truncate(0)} }
+    end
   rescue NotImplementedError
   end
 
-  def test_flock ## xxx
-    f = File.new(regular_file, "r+")
-    f.flock(File::LOCK_EX)
-    f.flock(File::LOCK_SH)
-    f.flock(File::LOCK_UN)
-    f.close
+  def test_flock_exclusive
+    File.open(regular_file, "r+") do |f|
+      f.flock(File::LOCK_EX)
+      assert_separately(["-rtimeout", "-", regular_file], "#{<<~begin}#{<<~"end;"}")
+      begin
+        open(ARGV[0], "r") do |f|
+          Timeout.timeout(0.1) do
+            assert(!f.flock(File::LOCK_SH|File::LOCK_NB))
+          end
+        end
+      end;
+      assert_separately(["-rtimeout", "-", regular_file], "#{<<~begin}#{<<~"end;"}")
+      begin
+        open(ARGV[0], "r") do |f|
+          assert_raise(Timeout::Error) do
+            Timeout.timeout(0.1) do
+              f.flock(File::LOCK_SH)
+            end
+          end
+        end
+      end;
+      f.flock(File::LOCK_UN)
+    end
+  rescue NotImplementedError
+  end
+
+  def test_flock_shared
+    File.open(regular_file, "r+") do |f|
+      f.flock(File::LOCK_SH)
+      assert_separately(["-rtimeout", "-", regular_file], "#{<<~begin}#{<<~"end;"}")
+      begin
+        open(ARGV[0], "r") do |f|
+          Timeout.timeout(0.1) do
+            assert(f.flock(File::LOCK_SH))
+          end
+        end
+      end;
+      assert_separately(["-rtimeout", "-", regular_file], "#{<<~begin}#{<<~"end;"}")
+      begin
+        open(ARGV[0], "r+") do |f|
+          assert_raise(Timeout::Error) do
+            Timeout.timeout(0.1) do
+              f.flock(File::LOCK_EX)
+            end
+          end
+        end
+      end;
+      f.flock(File::LOCK_UN)
+    end
   rescue NotImplementedError
   end
 
@@ -1187,11 +1557,7 @@ class TestFileExhaustive < Test::Unit::TestCase
       assert_integer_or_nil(fs1.rdev_minor)
       assert_integer(fs1.ino)
       assert_integer(fs1.mode)
-      unless /emx|mswin|mingw/ =~ RUBY_PLATFORM
-        # on Windows, nlink is always 1. but this behavior will be changed
-        # in the future.
-        assert_equal(hardlinkfile ? 2 : 1, fs1.nlink)
-      end
+      assert_equal(hardlinkfile ? 2 : 1, fs1.nlink)
       assert_integer(fs1.uid)
       assert_integer(fs1.gid)
       assert_equal(3, fs1.size)
@@ -1209,6 +1575,12 @@ class TestFileExhaustive < Test::Unit::TestCase
     end
   end
 
+  def test_stat_new_utf8
+    assert_nothing_raised do
+      File::Stat.new(utf8_file)
+    end
+  end
+
   def test_stat_ftype
     assert_equal("directory", File::Stat.new(@dir).ftype)
     assert_equal("file", File::Stat.new(regular_file).ftype)
@@ -1381,10 +1753,12 @@ class TestFileExhaustive < Test::Unit::TestCase
   end
 
   def test_size
-    assert_equal(3, File.open(regular_file) {|f| f.size })
-    File.open(regular_file, "a") do |f|
-      f.write("bar")
-      assert_equal(6, f.size)
+    [regular_file, utf8_file].each do |file|
+      assert_equal(3, File.open(file) {|f| f.size })
+      File.open(file, "a") do |f|
+        f.write("bar")
+        assert_equal(6, f.size)
+      end
     end
   end
 
